import os, pmm_config, plesk_config, errno, signal, time, logging
from backend_interface import *
import execution_result
from threading import Thread
from subproc import CmdLine, ExecuteException, NonzeroExitException
from threading import BoundedSemaphore
import agents_spec
import split_files
import streaming_extractor
from PacketServer import StopServer
import xml.parsers.expat
import api_rpc_messages
import time
from api_rpc_client import ApiRpcClient
from api_rpc_messages import *
import gzip_stream
from ReadLineAdapter import ReadLineAdapter
import operator

import agent_output
import agent_input

import shutil
import osutil


_logger = logging.getLogger("pmm.supervisor.sm")


def getAgentByName(name):
    try:
        as = agents_spec.AgentSpec(os.path.join(pmm_config.agents_dir(), name))
        return (DetectedPanel(agent = name, name = as["Description"]), None)
    except IOError, ioe:
        if ioe.errno == 2: #ENOENT
            return (None, "Unable to find agent " + name)
        return (None, str(e))
    except Exception, e:
        return (None, str(e))


class UnableToGetStatus(Exception):
    def __init__(self, res):
        self.res = res

class State(object):
    """
    State object represents the state in the supervisor statemachine.
    """

    def setSession(self, session):
        self.session = session

    def processRequest(self, action):
        if action.getGet_status():
            return self.getStatus()
        if action.getDelete_session():
            self.session.keepArchive = action.getDelete_session().getKeep_archive()
            self.stop()
            raise StopServer(ActionResponse(command_accepted = StatusCommandAccepted()))
        res = self.processSpecificRequest(action)
        if res:
            return res

        return (None, ActionResponse(command_parse_error = \
                                     StatusCommandParseError("Unexpected action in request")))

    def processSpecificRequest(self, action):
        pass

    def stop(self):
        pass

    def do(self):
        t = Thread(target = self.doActivityRunner)
        t.setDaemon(True)
        t.start()

    def doActivityRunner(self):
        pass

    def getStatus(self):
        try:
            return (None, self.realGetStatus())
        except UnableToGetStatus, e:
            ns = FinishedState(e.res)
            return (ns, ns.realGetStatus())
        except Exception:
            self.stop()
            stack = stacktrace.stacktrace()
            _logger.debug(stack)
            ns = FinishedState(StatusFinished("error",
                    [MessageType("Error in " + self.__class__.__name__ + ".getStatus()",
                                "", "", "", stack)]))
            return (ns, ns.realGetStatus())

    def realGetStatus(self):
        class NotImplemented(Exception):
            pass
        raise NotImplemented

    def getResults(self):
        er = execution_result.ExecutionResult()
        er.build(minidom.parse(self.session.getResultsFileName()).childNodes[0])
        return er


class LongActionState(State):
    def __init__(self):
        self.actionStateMutex = BoundedSemaphore()
        self.wasStopped = False
        State.__init__(self)

    def doActivityRunner(self):
        try:
            newState = self.doActivity()
        except Exception:
            stack = stacktrace.stacktrace()
            msg = MessageType("Error in " + self.__class__.__name__ + ".doActivity()","", "", "", stack)
            newState = FinishedState(StatusFinished("error", [msg]))

        self.actionStateMutex.acquire()
        try:
            if not self.wasStopped:
                self.session.setState(newState)
        finally:
            self.actionStateMutex.release()

    def stop(self):
        self.actionStateMutex.acquire()
        try:
            self.wasStopped = True
            self.stopAction()
        finally:
            self.actionStateMutex.release()

    def stopAction(self):
        pass


class BackendProcessState(LongActionState):
    def __init__(self):
        LongActionState.__init__(self)
        self.process = None

    def setProcess(self, process):
        if process == None:
            _logger.debug("Subprocess finished")
        else:
            _logger.debug("Subprocess " + str(process) + " encountered.")

        self.actionStateMutex.acquire()
        try:
            self.process = process
        finally:
            self.actionStateMutex.release()

    def stopAction(self):
        if self.process:
            self.process.kill(signal.SIGTERM)
        else:
            _logger.debug("No subprocess is active during manual stop")

    def realGetStatus(self):
        statusFileName = self.session.sessionDirectory + "/migration.status"
        try:
            statusFile = open(statusFileName)
        except IOError, e:
            # check for non-existing file
            if e.errno != errno.ENOENT:
                raise

            # file does not exist, so assume backend didn't create it yet
            return ActionResponse(working = StatusWorking(starting = StatusWorkingStarting()))
        else:
            try:
                status = StatusWorking()
                content = statusFile.read()
                _logger.debug("Status:\n" + content)
                d = minidom.parseString(content)
                status.build(d.childNodes[0])
                return ActionResponse(working = status)
            finally:
                statusFile.close()


class AwaitingStartState(State):
    def processSpecificRequest(self, action):
        if action.getStart_migration():
            self.session.connectivityInfo = action.getStart_migration().getConnectivity_info()
            return (SendScoutState(), ActionResponse(command_accepted = StatusCommandAccepted()))

        if action.getStart_import_migration():
            return (AwaitingImportParameters(), \
                    ActionResponse(command_accepted = StatusCommandAccepted()))

        if action.getStart_api_migration():
            sam = action.getStart_api_migration()
            self.session.apiConnectivityInfo = sam.getConnectivity_info()
            self.session.apiRemoteAgentDir = sam.getRemote_agent_dir()
            self.session.selectedObjects = sam.getSelected_objects()
            return (ApiAgentState(), ActionResponse(command_accepted = StatusCommandAccepted()))

        if action.getStart_restore():
            sr = action.getStart_restore()
            self.session.file = sr.getFile()
            self.session.user = sr.getUser()
            self.session.userType = sr.getUser_type()
            return (RestoreBackupState(), ActionResponse(command_accepted = StatusCommandAccepted()))

    def realGetStatus(self):
        return ActionResponse(waiting = StatusAwaiting(StatusAwaitingStart()))


class SendScoutState(BackendProcessState):
    def doActivity(self):
        cmd = CmdLine(pmm_config.launchpad(), self)
        self.session.options(cmd, "Connectivity", "Session")
        cmd.arg("--send-scout")
        cmd.spawn()

        res = self.getResults()
        if res.getStatus() != "success":
            return FinishedState(res)

        return AwaitingAgentParameters()


class AwaitingAgentParameters(State):
    def __init__(self, wrongPlatform = None):
        self.wrongPlatform = wrongPlatform

    def processSpecificRequest(self, action):
        if action.getSet_agent_parameters():
            self.session.agentParameters = action.getSet_agent_parameters().getAgent_parameters()
            self.session.sessionParameters = action.getSet_agent_parameters().getSession_parameters()
            return (GetObjectsState(), ActionResponse(command_accepted = StatusCommandAccepted()))

    def realGetStatus(self):
        cmd = CmdLine(pmm_config.deployer())
        self.session.options(cmd, "Session")
        cmd.arg("--get-agents-list")
        out = cmd.spawn().stdout

        _logger.info(out)

        agents_list = AgentsList()
        agents_list.build(minidom.parseString(out.encode("UTF-8")).childNodes[0])

        scout_result = ScoutResult()
        scout_result.build(minidom.parse(self.session.sessionDirectory
                                         + "/scout-result.xml").childNodes[0])

        aap = StatusAwaitingAgentParameters(AgentInfo(agents_list,
                                                      scout_result.getOs_info(),
                                                      scout_result.getRemote_fs()),
                                            SessionInfo(scout_result.getDetected_panel(),
                                                        self.wrongPlatform))

        return ActionResponse(waiting = StatusAwaiting(awaiting_agent_parameters = aap))


class AwaitingImportParameters(State):
    def processSpecificRequest(self, action):
        if action.getSet_import_parameters():
            self.session.sessionParameters = action.getSet_import_parameters().getSession_parameters()
            self.session.agentParameters = AgentParameters(agent_name=self.session.detectPanel()[0])
            return (TranslateDumpInImportMigrationState(),
                    ActionResponse(command_accepted = StatusCommandAccepted()))

    def realGetStatus(self):
        (detectedPanel, wrongPlatform) = self.session.detectPanel()

        agent = None
        if detectedPanel:
            (agent, wrongPlatform) = getAgentByName(detectedPanel)

        aap = StatusAwaitingImportParameters(SessionInfo(agent, wrongPlatform))
        return ActionResponse(waiting = StatusAwaiting(awaiting_import_parameters = aap))


class RestoreBackupState(BackendProcessState):
    def doActivity(self):
        if self.session.file == "":
            return FinishedState(StatusFinished("error",
                                                [MessageType("Empty backup file name")]))

        backup_parts = split_files.findParts(self.session.file)
        backup_cat_cmd = split_files.open(backup_parts)

        if gzip_stream.isGzipped(self.session.file):
            backup_stream = gzip_stream.GzipReadStream(backup_cat_cmd.stdout)
            backup_stream = ReadLineAdapter(backup_stream)
        else:
            backup_stream = backup_cat_cmd.stdout

        streaming_extractor.StreamingExtractor(backup_stream,
                                               self.session.sessionDirectory,
                                               self.session.getConvertedDumpFileName())

        return AwaitingRestoreObjectsState()

class AgentDeleter(object):
    def __init__(self, session):
        self.session = session

    def __call__(self):
        cmd = CmdLine(pmm_config.launchpad())
        self.session.options(cmd, "Connectivity", "Session", "Agent", "AgentDir")
        cmd.arg("--remove-agent")

        try:
            cmd.spawn()
        except ExecuteException, e:
            pass

class AgentDeleterMixin(object):
    def stop(self):
        t = Thread(target = AgentDeleter(self.session))
        t.start()
        LongActionState.stop(self)

class GetObjectsState(AgentDeleterMixin, BackendProcessState):
    def __init__(self):
        BackendProcessState.__init__(self)

    def doActivity(self):

        cmd = CmdLine(pmm_config.launchpad(), self)
        self.session.options(cmd, "Connectivity", "Session", "Agent", "AgentDir")
        cmd.arg("--deploy-agent")

        try:
            stdout = cmd.spawn().stdout
        except ExecuteException, e:
            return AwaitingAgentParameters(e.subprocess.stdout)

        res = self.getResults()
        if res.getStatus() != "success":
            return AwaitingAgentParameters(stdout)

        cmd = CmdLine(pmm_config.launchpad(), self)
        self.session.options(cmd, "Connectivity", "Session", "Agent", "AgentDir", "Dump")
        cmd.arg("--no-content")
        cmd.arg("--dump-all")
        cmd.spawn()

        res = self.getResults()
        if res.getStatus() != "success":
            return FinishedState(res)

        return TranslateDumpInRegularMigrationState()


class TranslateDumpState(BackendProcessState):
    def translateDump(self):
        cmd = CmdLine(pmm_config.deployer(), self)
        self.session.options(cmd, "Session", "Agent")
        cmd.arg("--translate-dump")
        cmd.spawn()

        res = self.getResults()
        if res.getStatus() != "success":
            return FinishedState(res)


class TranslateDumpInRegularMigrationState(AgentDeleterMixin, TranslateDumpState):
    def doActivity(self):
        state = self.translateDump()
        if state:
            return state
        return AwaitingAgentObjectsState()


class TranslateDumpInImportMigrationState(TranslateDumpState):
    def doActivity(self):
        state = self.translateDump()
        if state:
            return state
        return AwaitingImportObjectsState()


class AwaitingObjectsState(State):
    def realGetStatus(self):
        cmd = CmdLine(pmm_config.deployer())
        self.session.options(cmd, "Session")
        cmd.arg("--get-accounts")
        out = cmd.spawn().stdout

        res = self.getResults()
        if res.getStatus() != "success":
            raise UnableToGetStatus(res)

        ao = StatusAwaitingObjects()
        ao.build(minidom.parseString(out.encode("UTF-8")).childNodes[0])

        return ActionResponse(waiting = StatusAwaiting(awaiting_objects = ao))

    def setMigrationObjects(self, request):
        self.session.selectedObjects = request.getSet_migration_objects().getSelected_objects()


class AwaitingAgentObjectsState(AwaitingObjectsState):
    def processSpecificRequest(self, action):
        if action.getSet_migration_objects():
            self.setMigrationObjects(action)

            # Remove dumping-status.xml to avoid reusing old dumping-status.xml in getRunningAgentState()
            statusFileName = os.path.join(self.session.sessionDirectory, "dumping-status.xml")
            osutil.unlink_nothrow(statusFileName)

            return (DumpSelectedState(), ActionResponse(command_accepted = StatusCommandAccepted()))


class AwaitingImportObjectsState(AwaitingObjectsState):
    def processSpecificRequest(self, action):
        if action.getSet_migration_objects():
            self.setMigrationObjects(action)
            return (ImportDeploymentState(), ActionResponse(command_accepted = StatusCommandAccepted()))


class AwaitingRestoreObjectsState(AwaitingObjectsState):
    def processSpecificRequest(self, action):
        if action.getSet_migration_objects():
            self.setMigrationObjects(action)
            return (RestoreDeploymentState(), \
                    ActionResponse(command_accepted = StatusCommandAccepted()))

def getRunningAgentState(session):
    statusFileName = os.path.join(session.sessionDirectory, "dumping-status.xml")
    try:
        statusFile = open(statusFileName)
    except IOError, e:
        # check for non-existing file
        if e.errno != errno.ENOENT:
            raise
        # file does not exist, so assume backend didn't create it yet
        s = StatusWorkingStarting()
        return ActionResponse(working = StatusWorking(starting = s))
    else:
        try:
            content = statusFile.read()
            _logger.debug("Agent dumping status:\n" + content)

            status = AgentDumpingStatus()
            d = minidom.parseString(content)
            status.build(d.childNodes[0])

            s = MigrationAgentStatusWorkingDumping()
            totalDomains = totalClients = 0

            so = session.selectedObjects
            if so.getMigrate_domains():
                totalDomains = len(so.getMigrate_domains()[0].getDomain())
                totalClients = 0
            elif so.getMigrate_clients():
                totalClients = len(so.getMigrate_clients().getClient())
                domainNums = [len(c.getDomain()) for c in so.getMigrate_clients().getClient()]
                totalDomains = reduce(operator.add, domainNums, 0)
            elif so.getMigrate_server():
                totalClients = len(so.getMigrate_server().getClient())
                domainNums = [len(c.getDomain()) for c in so.getMigrate_server().getClient()]
                totalDomains = reduce(operator.add, domainNums, 0)

            s.setTotal_domains(totalDomains)
            s.setTotal_clients(totalClients)
            s.setCompleted_domains(totalDomains - status.getDomains_left())
            s.setCompleted_clients(totalClients - status.getAccounts_left())

            return ActionResponse(working = StatusWorking(dumping = s))
        finally:
            statusFile.close()


class DumpSelectedState(AgentDeleterMixin, BackendProcessState):
    def doActivity(self):
        cmd = CmdLine(pmm_config.launchpad(), self)
        self.session.options(cmd, "Connectivity", "Session", "Agent", \
                             "AgentDir", "Dump", "DumpMigratingObjects")
        cmd.spawn()

        res = self.getResults()
        if res.getStatus() != "success":
            return FinishedState(res)

        return DownloadState()

    # FIXME: dumping-status.xml needs to be merged into migration.status or even completely eleminated
    def realGetStatus(self):
        return getRunningAgentState(self.session)


class DownloadState(AgentDeleterMixin, BackendProcessState):
    def doActivity(self):

        cmd = CmdLine(pmm_config.deployer(), self)
        self.session.options(cmd, "Session", "Agent")
        cmd.arg("--translate-dump")
        cmd.spawn()

        res = self.getResults()
        if res.getStatus() != "success":
            return FinishedState(res)

        cmd = CmdLine(pmm_config.launchpad(), self)
        self.session.options(cmd, "Connectivity", "Session", "Agent", \
                             "AgentDir", "Dump", "GetContent")
        cmd.arg("--get-content")
        cmd.spawn()

        res = self.getResults()
        if res.getStatus() != "success":
            return FinishedState(res)

        return DeploymentState()


class RpcError(Exception):
    def __init__(self, code, message):
        Exception.__init__(self)
        self.code = code
        self.message = message


def checkStatus(result, request):
    if result.getStatus() != 'ok':
        raise RpcError(ApiAgentState.AERR_GENERAL, apiRpcErrorMessage(request, result))


class ApiResultState(object):
    def __init__(self, totalSteps = 0, currentStep = 0):
        self.__stateMutex = BoundedSemaphore()
        self.__totalSteps = totalSteps
        self.__currentStep = currentStep

    def setCurrentStep(self, step):
        self.__stateMutex.acquire()
        try:
            self.__currentStep = step
        finally:
            self.__stateMutex.release()

    def setTotalSteps(self, step):
        self.__stateMutex.acquire()
        try:
            self.__totalSteps = step
        finally:
            self.__stateMutex.release()

    def realGetStatus(self):
        self.__stateMutex.acquire()
        try:
            return ActionResponse(api_working = (MigrationApiStatusWorking(
                        total = self.__totalSteps,
                        completed = self.__currentStep)))
        finally:
            self.__stateMutex.release()


class ApiAgentState(ApiResultState, LongActionState):
    agent_AERR_MIGRATION_CLIENT_LOGIN_NOT_VALID= "15012"
    agent_AERR_PROTO_NOTSUPP="1005"
    agent_AERR_COMPONENT_NOT_INSTALLED="1026"

    AERR_MIGRATION_INCOMPATIBLE_PROTOCOL="AERR_MIGRATION_INCOMPATIBLE_PROTOCOL"
    AERR_GENERAL="AERR_GENERAL"

    STATUS_REQUEST_TIMEOUT = 5

    def __init__(self):
        LongActionState.__init__(self)
        ApiResultState.__init__(self)

    def __doRequest(self, arc, req, answerClass):
        try:
            return self.__doRequestInternal(arc, req, answerClass)
        except RpcError:
            raise
        except xml.parsers.expat.ExpatError, e:
            raise RpcError(self.AERR_GENERAL, parsingErrorMessage(req, e, e.content))
        except Exception:
            raise RpcError(self.AERR_GENERAL, genericErrorMessage(req))

    def __doRequestInternal(self, arc, req, answerClass):
        _logger.info(req.get_data())
        return self.__doRealRequest(arc, req, answerClass)

    def __doRealRequest(self, arc, req, answerClass):
        try:
            resp = arc.doRequest(req)
            respContent = resp.read()
        except urllib2.HTTPError, e:
            if e.code == 404:
                raise RpcError(self.AERR_MIGRATION_INCOMPATIBLE_PROTOCOL,
                               "No API-RPC installed on source host")
            raise RpcError(self.AERR_GENERAL, httpErrorMessage(req, e))
        except urllib2.URLError, e:
            raise RpcError(self.AERR_GENERAL, urlErrorMessage(req, e))

        #FIXME add validation

        _logger.info("API-RPC response:" + respContent)

        answer = answerClass()
        try:
            answer.build(minidom.parseString(respContent).childNodes[0])
        except xml.parsers.expat.ExpatError, e:
            e.content = respContent
            raise

        if answer.getSystem():
            errCode = answer.getSystem().getErrcode()
            if errCode == self.agent_AERR_MIGRATION_CLIENT_LOGIN_NOT_VALID:
                raise RpcError(self.AERR_WRONGPASSWORD, answer.getSystem().getErrtext())
            elif errCode  == self.agent_AERR_PROTO_NOTSUPP:
                raise RpcError(self.AERR_MIGRATION_INCOMPATIBLE_PROTOCOL, answer.getSystem().getErrtext())
            else:
                raise RpcError(self.AERR_GENERAL, apiRpcErrorMessage(req, answer.getSystem()))

        return answer

    def doActivity(self):
        self.setTotalSteps(self.session.getDomainsCount() * 3 +  self.session.getClientsCount() * 3)
        try:
            aci = self.session.apiConnectivityInfo
            apiClient = ApiRpcClient(aci.getHost(), aci.getLogin(), aci.getPassword())

            #check protocol version
            get = agent_input.getType(gen_info = agent_input.NoneType())
            sot = agent_input.ServerOperatorType(get = [get])
            checkProtocolPacket = agent_input.RequestPacketType(server = [sot])
            checkProtocolRequest = apiClient.packetRequest(checkProtocolPacket)

            resp = self.__doRequest(apiClient, checkProtocolRequest, agent_output.ResponsePacketType)
            ## FIXME? no check for version?

            #check for installed migration API agent on source host
            checkAgent = agent_input.MigrationAgentInputType(check = agent_input.NoneType())
            checkAgentPacket = agent_input.RequestPacketType(migration_agent = [checkAgent])

            try:
                checkAgentRequest = apiClient.packetRequest(checkAgentPacket)
            except xml.parsers.expat.ExpatError, e:
                # 8.0.0 returns empty answer or garbage if no pmm installed
                raise RpcError(self.AERR_MIGRATION_INCOMPATIBLE_PROTOCOL,
                               "Migration manager is not installed on the source host")

            resp = self.__doRequest(apiClient, checkAgentRequest, agent_output.ResponsePacketType)
            answer = resp.getMigration_agent()[0].getCheck()[0]
            if answer.getResult().getStatus() != "ok":
                if answer.getResult().getErrcode() == self.agent_AERR_COMPONENT_NOT_INSTALLED:
                    raise RpcError(self.AERR_MIGRATION_INCOMPATIBLE_PROTOCOL, answer.getResult().getErrtext())
                else:
                    raise RpcError(self.AERR_GENERAL, apiRpcErrorMessage(checkAgentRequest, answer.getResult()))

            #upload agent file
            uploadRequest = apiClient.fileUploadRequest(file(pmm_config.plesk_agent_archive()))
            resp = self.__doRequest(apiClient, uploadRequest, agent_output.ResponsePacketType)
            result = resp.getUpload()[0].getResult()[0]

            checkStatus(result, uploadRequest)
            tmpFileName = result.getFile()

            #start migration
            arguments = agent_input.MigrationAgentInputStart(
                temporary_agent_file = tmpFileName,
                agent_dir = self.session.apiRemoteAgentDir,
                selected_objects = self.session.selectedObjects)
            startCommand = agent_input.MigrationAgentInputType(start = arguments)
            packet = agent_input.RequestPacketType(migration_agent = [startCommand])

            startRequest = apiClient.packetRequest(packet)
            resp = self.__doRequest(apiClient, startRequest,
                                    agent_output.ResponsePacketType)
            result = resp.getMigration_agent()[0].getStart()[0].getResult()

            checkStatus(result, startRequest)
            migrationId = result.getMigration_id()

            #while dumping, keep asking
            while True:
                arguments = agent_input.MigrationAgentInputGetStatus(migration_id = migrationId)
                get_status_request = \
                    agent_input.MigrationAgentInputType(get_status = arguments)
                packet = agent_input.RequestPacketType(migration_agent = [get_status_request])

                getStatusRequest = apiClient.packetRequest(packet)
                resp = self.__doRequest(apiClient, getStatusRequest,
                                        agent_output.ResponsePacketType)

                result = resp.getMigration_agent()[0].getGet_status()[0].getResult()
                checkStatus(result, getStatusRequest)

                if result.getFinished():
                    break

                working = result.getWorking()
                if not working.getPreparing():
                    dumping = working.getDumping()
                    step = dumping.getCompleted_domains() + dumping.getCompleted_clients()
                    self.setCurrentStep(step)

                time.sleep(self.STATUS_REQUEST_TIMEOUT)

            dumpUrl = result.getFinished().getDump_url()
            contentUrls = result.getFinished().getContent_url()
            password = result.getFinished().getPassword()

            dumpFile = file(self.session.getDumpFileName(), 'w')
            remoteFile = apiClient.openFile(dumpUrl, "migration", password)
            _logger.info("Downloading " + dumpUrl)
            shutil.copyfileobj(remoteFile, dumpFile)
            _logger.info(dumpUrl + " downloaded")

            if not os.path.exists(self.session.getArchivesDir()):
                os.mkdir(self.session.getArchivesDir())

            for contentUrl in contentUrls:
                _logger.info("Downloading " + contentUrl)
                contentFile = file(os.path.join(self.session.getArchivesDir(),
                                   os.path.basename(contentUrl)), 'w')
                remoteFile = apiClient.openFile(contentUrl, "migration", password)
                shutil.copyfileobj(remoteFile, contentFile)
                _logger.info(contentUrl + " downloaded")

            self.setCurrentStep(self.session.getDomainsCount() * 2 +
                                self.session.getClientsCount() * 2)

            packet = agent_input.RequestPacketType(
                migration_agent = [agent_input.MigrationAgentInputType(
                    finish = agent_input.MigrationAgentInputFinish(migrationId))])
            finishRequest = apiClient.packetRequest(packet)
            resp = self.__doRequest(apiClient, finishRequest,
                                    agent_output.ResponsePacketType)
            checkStatus(result, getStatusRequest)

            #FIXME: remove hardcoded 'mv'
            os.rename(self.session.getDumpFileName(), self.session.getPleskDumpFileName())

            return ApiDeploymentState(self.session.getDomainsCount() * 3
                                      + self.session.getClientsCount() * 3,
                                      self.session.getDomainsCount() * 2
                                      + self.session.getClientsCount() * 2)

        except RpcError, e:
            return FinishedState(StatusFinished("error",
                                                [MessageType(code = e.code, text = e.message)]))


class CommonDeploymentState(BackendProcessState):
    def __init__(self, selectObjectsClass = None):
        BackendProcessState.__init__(self)
        self.selectObjectsClass = selectObjectsClass

    def doActivity(self):
        cmd = CmdLine(pmm_config.deployer(), self)
        self.session.options(cmd, "Session")
        self.session.getIpMap(cmd)
        cmd.arg("--deploy-dump")

        self.passSpecificDeploymentOptions(cmd)

        try:
            cmd.spawn()
            return FinishedState(self.getResults(), self.selectObjectsClass)
        except ExecuteException, e:
            if (os.access(self.session.getResultsFileName(), os.R_OK)):
                return FinishedState(self.getResults(), self.selectObjectsClass)
            raise

    def passSpecificDeploymentOptions(self, cmd):
        pass


class ApiDeploymentState(ApiResultState, CommonDeploymentState):
    def __init__(self, totalSteps, currentStep):
        ApiResultState.__init__(self, totalSteps, currentStep)
        CommonDeploymentState.__init__(self)


class DeploymentState(AgentDeleterMixin, CommonDeploymentState):
    def __init__(self):
        CommonDeploymentState.__init__(self, GetObjectsState)
    def passSpecificDeploymentOptions(self, cmd):
        self.session.options(cmd, "Deploy")

class ImportDeploymentState(AgentDeleterMixin, CommonDeploymentState):
    def __init__(self):
        CommonDeploymentState.__init__(self, TranslateDumpInImportMigrationState)
    def passSpecificDeploymentOptions(self, cmd):
        self.session.options(cmd, "Deploy")


class RestoreDeploymentState(CommonDeploymentState):
    def __init__(self):
        CommonDeploymentState.__init__(self, AwaitingRestoreObjectsState)

    def passSpecificDeploymentOptions(self, cmd):
        cmd.arg("--backup-user=" + self.session.user)
        cmd.arg("--backup-user-type=" + self.session.userType)
        cmd.arg("--backup-file=" + self.session.file)


class FinishedState(State):
    def __init__(self, status, selectObjectsClass = None):
        self.status = status
        self.selectObjectsClass = selectObjectsClass

    def realGetStatus(self):
        return ActionResponse(finished = self.status)

    def processSpecificRequest(self, action):
        if self.selectObjectsClass \
                and action.getGo_to_select_objects():
            del self.session.selectedObjects
            return (self.selectObjectsClass(),
                    ActionResponse(command_accepted = StatusCommandAccepted()))

    def stop(self):
        try:
            AgentDeleter(self.session)()
        except AttributeError:
            pass
